Require Import ssreflect.

(* frf_ssr.v : 周波数応答に関する exp 関数の微分 *)
Section frf_ssr.

  (* 複素数体のつもり *)
  Variables C : Type.
  Variables one_c mone_c zero_c : C.
  Variable add_c mult_c : C->C->C.

  (* 複素数としての加法, 乗法 *)
  Notation "a + b" := (add_c a b).
  Notation "a * b" := (mult_c a b).

  
  Axiom one_mult_c : forall (c : C), one_c * c = c.
  Axiom zero_add_c : forall (c : C), zero_c + c = c.
  Axiom munit_c : mone_c + one_c = zero_c.
  Axiom assoc_mult_c : forall (c1 c2 c3 : C), (c1 * c2) * c3 = c1 * (c2 * c3).
  Axiom comm_mult_c : forall (c1 c2 : C), c1 * c2 = c2 * c1.
  Axiom assoc_add_c : forall (c1 c2 c3 : C), (c1 + c2) + c3 = c1 + (c2 + c3).
  Axiom comm_add_c : forall (c1 c2 : C), c1 + c2 = c2 + c1.
  Axiom dist_c : forall (c1 c2 c : C), (c1 + c2) * c = c1 * c + c2 * c.
  Parameter mult_zero_c : forall (c : C), zero_c * c = zero_c.


  (* 関数としての 1, 0 *)
  Definition add (f g : C->C) (z : C) := f z + g z. 
  Definition mult (f g : C->C) (z : C) := f z * g z.
  Definition cmult (c : C) (f : C->C) (z : C) := c * f z. 
  Definition one (z : C) : C := one_c.
  Definition zero (z : C) : C := zero_c.

  Lemma cmult_one : forall (f : C->C) (z : C), (cmult one_c f) z = f z.   
  Proof.
    move => f z.
    rewrite /cmult.
    rewrite one_mult_c.
    by [].
  Qed.

  Lemma cmult_zero : forall (f : C->C) (z : C), (cmult zero_c f) z = zero z.
  Proof.
    move => f z.
    rewrite /cmult.
    rewrite /zero.
    rewrite mult_zero_c.
    by [].
  Qed.
    
  (* 1(z) * f(z) = f(z) : ssreflect 使用時 10点 *)
  Lemma mult_one : forall (f : C->C) (z : C), (mult one f) z = f z.
  Admitted.

  (* f(z) + g(z) = g(z) + f(z) : ssreflect 使用時 10点 *)
  Lemma comm_mult : forall (f g : C->C) (z : C), (mult f g) z = (mult g f) z.
  Proof.
  Admitted.


  (* ((f+g)+h)(z) = (f+(g+h))(z) : ssreflect 使用時 10点 *)
  Lemma assoc_mult : forall (f g h : C->C) (z : C), (mult (mult f g) h) z = (mult f (mult g h)) z.
  Admitted.

  
  Variable pow : C->C->C.
  Variable exp : C -> C.
  (* 指数法則 *)
  Axiom exp_add_mult : forall (z1 z2 : C), exp (z1 + z2) = (exp z1) * (exp z2).
  Axiom exp_pow : forall (a z : C), exp (a*z) = pow a (exp z).
    
  (* 微分 *)
  Variable diff : (C->C) -> (C->C).
  Variable diff_add : forall (f g : C->C) (z : C), (diff (add f g)) z = (add ((diff f)) ((diff g))) z.
  Variable diff_cmult : forall (c : C) (f : C->C) (z : C), (diff (cmult c f)) z = (cmult c (diff f)) z.
  Axiom diff_exp : forall (z : C), (diff exp) z = exp z.
  Axiom diff_pow : forall (a w : C), diff (pow a) w= a * (pow (a+mone_c)) w.
  Definition composite (f g : C->C) (z : C) : C := f (g z).
  Lemma composite_cmult : forall (f g : C->C) (c z : C), (composite (cmult c f) g) z = c * (composite f g) z.
  Proof.
    move => f g c z.
    rewrite /composite.
    rewrite /cmult.
    by [].
  Qed.
  
  Axiom chain_rule : forall (f g : C->C) (z : C), (diff (composite f g)) z = ((composite (diff f) g) z) * (diff g) z. 

  
  Lemma exp_cz_1 : forall (c z : C), exp (c*z) = (composite (pow c) exp) z.
  Proof.
    move => c z.
    rewrite /composite.
    rewrite exp_pow.
    by [].
  Qed.

  (* exp (1*z) = exp z : ssreflect 使用時 10 点 *)
  Lemma exp_one_z : forall (z : C), exp (one_c*z) = exp z.
  Admitted.
    
  Lemma exp_cz_2 : forall (c z : C), exp ((c+mone_c)*z) * exp (one_c*z) = exp (c*z).
  Proof.
    move => c z.
    rewrite -exp_add_mult.
    rewrite -dist_c.
    have Hz : (c + mone_c + one_c = c).
    - rewrite assoc_add_c.
      rewrite munit_c.
      have Hcm : (c + zero_c = zero_c + c).
      + rewrite comm_add_c.
          by [].
      + rewrite Hcm.
        rewrite zero_add_c.
        by [].
    rewrite Hz.
    by [].
  Qed.
    
    
  (*
    (exp (cz))' に合成関数の微分方程式を適用して
    (exp (cz))' = c (exp (z))^(c-1) exp z
    が成り立つことを証明
   *)
  Lemma diff_pow_exp : forall (c z : C), (diff (composite (pow c) exp)) z = (c * ((composite (pow (c + mone_c)) exp) z)) * exp z.
  Proof.
    move => c z.
    rewrite chain_rule.
    rewrite diff_exp.
    have Hc : (composite (diff (pow c)) exp z = c * (composite (pow (c+mone_c)) exp z)).
    - rewrite /composite.
      rewrite diff_pow.
      by [].
    - rewrite Hc.
        by [].
  Qed.

  Definition exp_c (c : C) : C->C := composite (pow c) exp.
  Theorem diff_exp_cz : forall (c z : C), diff (exp_c c) z = c * (exp_c c z).
  Proof.
    move => c z.
    rewrite /exp_c.
    have Hc: (composite (pow c) exp z = (composite (pow (c + mone_c)) exp) z * exp z).
    - repeat rewrite -exp_cz_1.
      rewrite -exp_cz_2.
      rewrite exp_one_z.
      by [].
    rewrite diff_pow_exp.
    rewrite Hc.
    rewrite -assoc_mult_c.
    by [].
  Qed.

  Definition d_exp_c (c : C) : C->C := (cmult c (exp_c c)).
  Theorem diff_1st : forall (c z : C), d_exp_c c z = diff (exp_c c) z.
  Proof.
    move => c z.
    rewrite /d_exp_c.
    rewrite /cmult.
    rewrite -diff_exp_cz.
    by [].
  Qed.
  
  Lemma diff_d_exp : forall (c z : C), diff (d_exp_c c) z = c * (c * exp_c c z).  
  Proof.
    move => c z.
    rewrite /d_exp_c.
    rewrite diff_cmult.
    rewrite /cmult.
    rewrite diff_exp_cz.
    by [].
  Qed.

  Definition d2_exp_c (c : C) : C->C := (cmult c (d_exp_c c)).

  (* 15 点問題 *)
  Theorem diff_2nd : forall (c z : C), d2_exp_c c z = diff (d_exp_c c) z.
  Admitted.
  
End frf_ssr.