
(*
  An Example of Finete Element Method and Newmark-beta method by OCaml
  un exemple de méthode des elements finis et méthode de Newmark par OCaml
  OCaml で書いた有限要素法 + Newmark-β 法の計算例

  [ Build 時の注意 ]
  Build 時は opam, ocamlfind, Lacaml が必要になります.
  ocamlfind ocamlopt -o rotor_dynamics -package lacaml -linkpkg lacaml.cmxa rotor_dynamics.ml
*)

(* 
   この有限要素法プログラムは
   名古屋大学大学院 工学研究科 機械理工学専攻 井上剛志教授
   が作成されたプログラムをもとに作成しています.
 *)
open Format
open Bigarray

open Lacaml.D
open Lacaml.Io

  
let pi = 4. *. atan 1.

(* べき乗 *)
let rec pow a x n =
  if n >= 1 then pow (a*.x) x (n-1)
  else a

let ( *^ ) a n = pow 1. a n
  
(* 横ベクトルの計算 *)
let plus_row u v =
  let dim = Array.length u in
  Array.init dim (fun n -> u.(n) +. v.(n))

let minus_row u v =
  let dim = Array.length u in
  Array.init dim (fun n -> u.(n) -. v.(n))

let prod_row c v =
  let dim = Array.length v in
  Array.init dim (fun n -> c *. v.(n))

(* 行列の計算 *)
let ( +! ) m1 m2 =
  let dim = Array.length m1 in
  Array.init dim (fun n -> plus_row m1.(n) m2.(n))

let ( -! ) m1 m2 =
  let dim = Array.length m2 in
  Array.init dim (fun n -> minus_row m1.(n) m2.(n))

let ( *< ) c m =
  let dim = Array.length m in
  Array.init dim (fun n -> prod_row c m.(n))

let bilinear u v i j =
  let l = Array.length u.(0) in
  let uv = Array.init l (fun n -> u.(i).(n) *. v.(n).(j)) in
  Array.fold_left (+.) 0. uv

(* 行列の積 *)
let ( *! ) a b =
  let row = Array.length a in
  let col = Array.length b.(0) in
  let ab_row k = Array.init col (fun n -> bilinear a b k n) in
  Array.init row ab_row

(* クロネッカーのデルタ *)
let kronecker i j = if i = j then 1. else 0.

(* トレース *)
let tr m =
  let row = Array.length m in
  let col = Array.length m.(0) in
  let tr_row k = Array.init row (fun i -> m.(i).(k)) in
  Array.init col tr_row

(* パラメーター *)
let nfem = 5
let nr = 2
let ns = 2
let g = 9.81e3
let omega_r = 2. *. pi *. 200.


(* 密度, ヤング率 *)
(* 軸 *)
let rho_axis = 7.86e-6 (* 密度 *)
let e_axis = 1.91446264e8 (* Young 率 *)
    
let r_axis = 15. (* 軸の半径 *)
let dmu_axis = rho_axis *. (r_axis *^ 2) *. pi (* 軸の線密度 *)
let l_axis = 200. (* 軸の長さ *)
let cg_axis = 0.0302939 (* 減衰係数 *)
let dip_axis = dmu_axis *. (r_axis *^ 2) /. 2. (* 極慣性モーメントの線密度 *)
let dia_axis = ((2. *. r_axis) *^ 4) *. pi /. 64. (* 軸の断面 2 次モーメント *)

(* 円板 *)
let rho_disk = 7.93e-6 (* 密度 *)
let r_disk = 150. (* 半径 *)
let h_disk = 30. (* 厚さ *)
let m_disk = ((r_disk *^ 2) -. (r_axis *^ 2)) *. h_disk *. rho_disk *. pi
let i_disk = m_disk *. ((r_disk *^ 2) /. 4. +. (h_disk *^ 2) /. 12.)
let ip_disk = m_disk *. (r_disk *^ 2) /. 2.
let tau = 0.

(* 不釣り合い *)
let eu = 0.1
let alpha = 0.

(* 軸受けの剛性 *)
let k_bearing_x = 2.28e6
let k_bearing_y = 2.28e6


(* 要素質量マトリックス *)
let matrix_mt_base l =
  let mt_base_cmp n m =
    match (n+1,m+1) with
    (*  *)
    | (1,1) ->  156.
    | (1,4) ->  22. *. l
    | (1,5) ->  54.
    | (1,8) ->  -13. *. l
    (*  *)
    | (2,2) ->  156.
    | (2,3) -> -22. *. l
    | (2,6) ->  54.
    | (2,7) ->  13. *. l
    (*  *)
    | (3,2) -> -22. *. l
    | (3,3) ->  4. *. l *. l
    | (3,6) -> -13. *. l
    | (3,7) -> -3. *. l *. l
    (*  *)
    | (4,1) ->  22. *. l
    | (4,4) ->  4. *. l *. l
    | (4,5) ->  13. *. l
    | (4,8) -> -3. *. l *. l
    (*  *)
    | (5,1) ->  54.
    | (5,4) ->  13. *. l
    | (5,5) ->  156.
    | (5,8) -> -22. *. l
    (* *)
    | (6,2) ->  54.
    | (6,3) -> -13. *. l
    | (6,6) ->  156.
    | (6,7) ->  22. *. l
    (* *)
    | (7,2) ->  13. *. l
    | (7,3) -> -3. *. l *. l
    | (7,6) ->  22. *. l
    | (7,7) ->  4. *. l *. l
    (* *)
    | (8,1) -> -13. *. l
    | (8,4) -> -3. *. l *. l
    | (8,5) -> -22. *. l
    | (8,8) ->  4. *. l *. l
    (* *)
    | (_,_) -> 0.0
  in
  Array.init 8 (fun r -> Array.init 8 (mt_base_cmp r))

let matrix_mt mu l =
  (mu *. l /. 420.) *< matrix_mt_base l

(* *)
let matrix_mr_base l =
  let mr_base_cmp n m =
    match (n+1,m+1) with
    (*  *)
    | (1,1) ->  36.
    | (1,4) ->  3. *. l
    | (1,5) -> -36.
    | (1,8) ->  3. *. l
    (*   *)
    | (2,2) ->  36.
    | (2,3) -> -3. *. l
    | (2,6) -> -36.
    | (2,7) -> -3. *. l
    (*   *)
    | (3,2) -> -3. *. l
    | (3,3) ->  4. *. l *. l
    | (3,6) ->  3. *. l
    | (3,7) -> -. l *. l
    (*   *)
    | (4,1) ->  3. *. l
    | (4,4) ->  4. *. l *. l
    | (4,5) -> -3. *. l
    | (4,8) -> -. l *. l
    (*   *)
    | (5,1) -> -36.
    | (5,4) -> -3. *. l
    | (5,5) ->  36.
    | (5,8) -> -3. *. l
    (*   *)
    | (6,2) -> -36.
    | (6,3) ->  3. *. l
    | (6,6) ->  36.
    | (6,7) ->  3. *. l
    (* *)
    | (7,2) -> -3. *. l
    | (7,3) -> -. l *. l
    | (7,6) ->  3. *. l
    | (7,7) ->  4. *. l *. l
    (* *)
    | (8,1) ->  3. *. l
    | (8,4) -> -. l *. l
    | (8,5) -> -3. *. l
    | (8,8) ->  4. *. l *. l
    (* *)
    | (_,_) -> 0.0
  in
  Array.init 8 (fun r -> Array.init 8 (mr_base_cmp r))

let matrix_mr mu r l=
  (mu *. r *. r /. (120. *. l)) *< (matrix_mr_base l) 

(* 要素質量マトリックス *)
let matrix_m mu r l =
  (matrix_mt mu l) +! (matrix_mr mu r l)

(* 要素減衰マトリックス *)
let matrix_c c l =
  (c *. l /. 420.) *< (matrix_mt_base l)

(* 要素ジャイロマトリックス *)
let matrix_g_base l =
  let g_base_cmp n m =
    match (n+1,m+1) with
    (*  *)
    | (1,2) ->   36.
    | (1,3) ->  -3. *. l
    | (1,6) ->  -36.
    | (1,7) ->  -3. *. l
    (*  *)
    | (2,1) ->  -36.
    | (2,4) ->  -3. *. l
    | (2,5) ->   36.
    | (2,8) ->  -3. *. l
    (*  *)
    | (3,1) ->   3. *. l
    | (3,4) ->   4. *. l *. l
    | (3,5) ->  -3. *. l
    | (3,8) ->  -. l *. l
    (*  *)
    | (4,2) ->   3. *. l
    | (4,3) ->  -4. *. l *. l
    | (4,6) ->  -3. *. l
    | (4,7) ->   l *. l
    (*  *)
    | (5,2) ->  -36.
    | (5,3) ->   3. *. l
    | (5,6) ->   36.
    | (5,7) ->   3. *. l
    (* *)
    | (6,1) ->   36.
    | (6,4) ->   3. *. l
    | (6,5) ->  -36.
    | (6,8) ->   3. *. l
    (* *)
    | (7,1) ->   3. *. l
    | (7,4) ->  -. l *. l
    | (7,5) ->  -3. *. l
    | (7,8) ->   4. *. l *. l
    (* *)
    | (8,2) ->   3. *. l
    | (8,3) ->   l *. l
    | (8,6) ->  -3. *. l
    | (8,7) ->  -4. *. l *. l
    (* *)
    | (_,_) -> 0.0
  in
  Array.init 8 (fun r -> Array.init 8 (g_base_cmp r))

let matrix_g omega ip l =
  (omega *. ip /. (30.0 *. l)) *< matrix_g_base l

let matrix_cg c omega ip l =
  (matrix_c c l) +! (matrix_g omega ip l)

(* 要素剛性マトリックス *)
let matrix_k_base l =
  let k_base_cmp n m =
    match (n+1,m+1) with
    (*  *)
    | (1,1) ->   12.
    | (1,4) ->   6. *. l
    | (1,5) ->  -12.
    | (1,8) ->   6. *. l
    (*  *)
    | (2,2) ->   12.
    | (2,3) ->  -6. *. l
    | (2,6) ->  -12.
    | (2,7) ->  -6. *. l
    (*  *)
    | (3,2) ->  -6. *. l
    | (3,3) ->   4. *. l *. l
    | (3,6) ->   6. *. l
    | (3,7) ->   2. *. l *. l
    (*  *)
    | (4,1) ->   6. *. l
    | (4,4) ->   4. *. l *. l
    | (4,5) ->  -6. *. l
    | (4,8) ->   2. *. l *. l
    (*  *)
    | (5,1) ->  -12.
    | (5,4) ->  -6. *. l
    | (5,5) ->   12.
    | (5,8) ->  -6. *. l
    (* *)
    | (6,2) ->  -12.
    | (6,3) ->   6. *. l
    | (6,6) ->   12.
    | (6,7) ->   6. *. l
    (* *)
    | (7,2) ->  -6. *. l
    | (7,3) ->   2. *. l *. l
    | (7,6) ->   6. *. l
    | (7,7) ->   4. *. l *. l
    (* *)
    | (8,1) ->   6. *. l
    | (8,4) ->   2. *. l *. l
    | (8,5) ->  -6. *. l
    | (8,8) ->   4. *. l *. l
    (* *)
    | (_,_) -> 0.0
  in
  Array.init 8 (fun r -> Array.init 8 (k_base_cmp r))

let matrix_k e ia l =
  (e *. ia /. (l *. l *. l)) *< (matrix_k_base l)


(*
  質量行列を計算
  軸のパラメータ付加 : matrix_mt, matrix_mr
  円板のパラメータ付加 : 
 *)
let m_sys =
  let lm = Array.make_matrix (4*nfem+4) (4*nfem+4) 0. in
  let lm_axis_elm = (matrix_mt dmu_axis l_axis) +! (matrix_mr dmu_axis r_axis l_axis) in
  let lm_axis = Array.make nfem lm_axis_elm in
  (* 軸のパラメータ付加 *)
  for n0 = 0 to nfem-1 do
    for n1 = 0 to 7 do
      for n2 = 0 to 7 do
	lm.(n1+4*n0).(n2+4*n0) <- lm.(n1+4*n0).(n2+4*n0) +. lm_axis.(n0).(n1).(n2);
      done;
    done;
  done;
  (* 円板のパラメータ付加 *)
  lm.(4*nr).(4*nr) <- lm.(4*nr).(4*nr) +. m_disk;
  lm.(4*nr+1).(4*nr+1) <- lm.(4*nr+1).(4*nr+1) +. m_disk;
  lm.(4*nr+2).(4*nr+2) <- lm.(4*nr+2).(4*nr+2) +. i_disk;
  lm.(4*nr+3).(4*nr+3) <- lm.(4*nr+3).(4*nr+3) +. i_disk;
  lm

(*
  matrix_g omega ip l
 *)
let cg_sys =
  let lcg = Array.make_matrix (4*nfem+4) (4*nfem+4) 0. in
  let lcg_axis_elem = matrix_cg cg_axis omega_r dip_axis l_axis in
  let lcg_axis = Array.make nfem lcg_axis_elem in
  (* *)
  for n0 = 0 to nfem-1 do
    for n1 = 0 to 7 do
      for n2 = 0 to 7 do
	lcg.(n1+4*n0).(n2+4*n0) <- lcg.(n1+4*n0).(n2+4*n0) +. lcg_axis.(n0).(n1).(n2);
      done;
    done;
  done;
  (* 円板のパラメータ付加 *)
  lcg.(4*nr+3).(4*nr+2) <- lcg.(4*nr+3).(4*nr+2) -. ip_disk *. omega_r;
  lcg.(4*nr+2).(4*nr+3) <- lcg.(4*nr+2).(4*nr+3) +. ip_disk *. omega_r;
  lcg
  
(*
  stiffness matrix
  剛性マトリックス
 *)
let k_sys =
  let lk = Array.make_matrix (4*nfem+4) (4*nfem+4) 0. in
  let lk_axis_elm = matrix_k e_axis dia_axis l_axis in
  let lk_axis = Array.make nfem lk_axis_elm in
  (* 軸のパラメータ付加 *)
  for n0 = 0 to 1 do
    for n1 = 0 to 7 do
      for n2 = 0 to 7 do
	lk.(n1+4*n0).(n2+4*n0) <- lk.(n1+4*n0).(n2+4*n0) +. lk_axis.(n0).(n1).(n2);
      done;
    done;
  done;
  (* 両端のベアリング剛性付加 *)
  lk.(0).(0) <- lk.(0).(0) +. k_bearing_x;
  lk.(4*nfem).(4*nfem) <- lk.(4*nfem).(4*nfem) +. k_bearing_x;
  lk.(1).(1) <- lk.(1).(1) +. k_bearing_y;
  lk.(4*nfem+1).(4*nfem+1) <- lk.(4*nfem+1).(4*nfem+1) +. k_bearing_y;
  lk



(* 連立方程式 : ax = b の解を計算する関数 *)
let solve a b =
  let b_mat = Mat.of_array b in
  gesv (Mat.of_array a) b_mat;
  Mat.to_array b_mat

(* OCaml 4.00 対応 *)
let (|>) x f = f x 
let scalar n u v =
  Array.length v |> (fun dim -> Array.init dim (fun j -> u.(n).(j) *. v.(j).(0)) |> Array.fold_left ( +. ) 0.)

let norm v =
  Array.map (fun x -> x.(0) *. x.(0)) v |> Array.fold_left (+.) 0.


let beta_nmb = 0.25
let gamma_nmb = 0.5
let n_smp = 1000000

let g_nmb n_deg (m,c,k) fn (x,v,a) =
  Array.init n_deg (fun id -> [| fn.(id).(0) -.  (scalar id m a) -. (scalar id c v) -. (scalar id k x) |])

(* n サンプルの値から n+1 サンプルの値を計算 *)
let newmark_beta_next n_deg (m, c, k, a_m) fn dt (x_p, v_p, a_p) =
  let a_c = a_p in
  let v_c = Array.init n_deg (fun id -> [| v_p.(id).(0) +. dt *. a_c.(id).(0) |]) in
  let x_c = Array.init n_deg (fun id -> [| x_p.(id).(0) +. dt *. v_c.(id).(0) +. 0.5 *. (dt *^ 2) *. a_c.(id).(0) |]) in
  (* 近似解の修正 *)
  let rec newmark_beta_correction (x,v,a) cnt =
    let delta_nmb = g_nmb n_deg (m,c,k) fn (x,v,a) in
    if norm delta_nmb > 1e-6 && cnt <= 500 then
      begin
	let da = solve a_m delta_nmb in
	let a' = Array.init n_deg (fun id -> [| a.(id).(0) +. da.(id).(0) |]) in
	let v' = Array.init n_deg (fun id -> [| v.(id).(0) +. dt *. gamma_nmb *. da.(id).(0) |]) in
	let x' = Array.init n_deg (fun id -> [| x.(id).(0) +. (dt *^ 2) *. beta_nmb *. da.(id).(0) |]) in
	newmark_beta_correction (x',v',a') (cnt+1)
      end
    else (x, v, a) 
  in

  (* 近似解の評価 *)
  newmark_beta_correction (x_c, v_c, a_c) 1
  

(* Newmark-β法での数値計算結果のプロット *)
let plot_newmark_beta ~n_plot ~m ~c ~k ~x0 ~v0 ~f_ext =
  let n_deg = Array.length x0 in
  let dt = 1. /. (float n_smp) in
  let a_m = m +! ((dt *. gamma_nmb)*< c) +! (((dt*^2) *. beta_nmb) *< k) in
  let rec plot_newmark_beta_at f (x_p, v_p, a_p) n_cur =
    let t = (float n_cur) /. (float n_smp) in
    let (x_c, v_c, a_c) = newmark_beta_next n_deg (m, c, k, a_m) (f t) dt (x_p, v_p, a_p) in

    print_string ((string_of_float t) ^ "\t");
    Array.iter (fun arr -> print_string ((string_of_float arr.(0)) ^ "\t")) x_c;
    print_endline "";
    if n_cur + 1 >= n_plot then () else plot_newmark_beta_at f (x_c, v_c, a_c) (n_cur + 1)
  in
  
  (* initial value *)
  let f0 = f_ext 0. in
  let b0 =
    Array.init n_deg (fun id -> [| f0.(id).(0) -. (scalar id c v0) -. (scalar id k x0) |]) in
  (* a0 を求める *)
  let a0 = solve m b0 in
  
  (* Newmark-β 法の計算 *)
  plot_newmark_beta_at f_ext (x0,v0,a0) 1


let () =
  let x0 = Array.make (4*nfem+4) [|0.|] in
  let v0 = Array.make (4*nfem+4) [|0.|] in

  let f_ub_elem n t =
    if n = 4*nr then [| m_disk *. eu *. omega_r *. omega_r *. (cos (omega_r *. t)) |]
    else if n = 4*nr+1 then [| m_disk *. eu *. omega_r *. omega_r *. (sin (omega_r *. t)) |]
    else [| 0. |]
  in

  let f_ub t = Array.init (4*nfem+4) (fun n -> f_ub_elem n t) in
  
  plot_newmark_beta ~n_plot:50000 ~m:m_sys ~c:cg_sys ~k:k_sys ~x0:x0 ~v0:v0 ~f_ext:f_ub
